//***************************************************************************
//
//   abstract class model (from MDS v.5.2)
//   translation mathematical model from math to C++
//   (c) 2002-2004 Dr.Bashnyakov A. 
//   (bash@unicyb.kiev.ua)
//
//***************************************************************************
//                                                                         
//   This program is free software; you can redistribute it and/or modify  
//   it under the terms of the GNU General Public License as published by  
//   the Free Software Foundation; either version 2 of the License, or     
//   (at your option) any later version.                                   
//                                                                         
//***************************************************************************


#ifndef _model_h_
#define _model_h_


#include "result.h"


enum _model_emulation_mode {EMULATION_DISABLED = 1, EMULATION_ENABLED = 2, EMULATION_RECORDING = 3};


// general class of model
template <class T = MDS_DEFAULT_TYPE, class EXCEPTION = MDS_DEFAULT_EXCEPTION>
class model 
{
public:
        explicit model(size_t dim = 0, int id = 0) :
                dimension(dim), identify(id), emulation_mode(EMULATION_DISABLED), emulation_order(1),
                x(dim), y(dim), phase(true, dim + 1)
        {
                        phase[0] = false;                
        }

        virtual ~model(){}

// right part of system: we NEED overload it, like that:
/*        virtual void RightPart(std::valarray<T>& f)
        {
                //equations (t is time; x[0], x[1],... - are phase):
                //f[0] = F0(t,x[0],x[1],...) - first equation;
                //f[1] = F1(t,x[0],x[1],...) - second equation...
        }  */
        virtual void RightPart(std::valarray<T>& f) = 0;

        std::valarray<T>& System(std::valarray<T>& ffunc, bool isEnlargedPhase = false)
        {
                MDS_ASSERT((isEnlargedPhase && ffunc.size() >= dimension + 1) || (!isEnlargedPhase && ffunc.size() >= dimension), EXCEPTION,
                        "[model] insufficient size of valarray in System(valarray<T>&, bool)");

                size_t cnt;

                if (emulation_mode == EMULATION_ENABLED)
                        emulation_table.GetApproximation(y, t, emulation_order);
                else
                {
                        RightPart(y);
                        
                        if (emulation_mode == EMULATION_RECORDING) 
                                emulation_table.Add(t, y, true);
                }
                
                if (isEnlargedPhase) 
                {
                        for (cnt = 1; cnt <= dimension; cnt++) ffunc[cnt] = y[cnt - 1];

                        ffunc[0] = t;
                }
                else  
                        for (cnt = 0; cnt < dimension; cnt++) ffunc[cnt] = y[cnt];

                return ffunc;
        }

        T& System(T& time, std::valarray<T>& func)
        {
                MDS_ASSERT(func.size() >= dimension, EXCEPTION,
                        "[model] valarray size is too low in System(T&, valarray<T>&)");

                size_t cnt;

                std::valarray<T> ffunc(dimension);

                System(ffunc);

                time = t;

                for (cnt = 0; cnt < dimension; cnt++) 
                        func[cnt] = ffunc[cnt];

                return time;
        }

        std::valarray<T> System(bool isEnlargedPhase = false)
        {
                size_t dim = dimension;
                if (isEnlargedPhase) dim++;

                std::valarray<T> ffunc(dim);
                
                System(ffunc, isEnlargedPhase);
                
                return ffunc;
        }

        static std::valarray<T>& System(const std::valarray<model<T,EXCEPTION> *>& pModels, 
                std::valarray<T>& ffunc, bool isEnlargedPhase = false) 
        { 
                size_t sysCnt, modCnt, modDim, shift = GetDimension(pModels), sysDim = pModels.size();

                std::valarray<T> buf(shift + 1);
                
                MDS_ASSERT((isEnlargedPhase && ffunc.size() >= shift + 1) || (!isEnlargedPhase && ffunc.size() >= shift), EXCEPTION,
                        "[model] insufficient size of valarray (argument 2) in System(const valarray<model<T,EXCEPTION> *>&, valarray<T>&, bool)");

                shift = 0;

                if (isEnlargedPhase)
                {
                        for (sysCnt = 0; sysCnt < sysDim; sysCnt++)
                        {
                                modDim = (pModels[sysCnt])->dimension;

                                (pModels[sysCnt])->System(buf, true);
                                
                                for (modCnt = 1; modCnt <= modDim; modCnt++)
                                        ffunc[modCnt + shift] = buf[modCnt];
                                
                                shift += modDim;
                        }

                        ffunc[0] = buf[0];
                }
                else
                {
                        for (sysCnt = 0; sysCnt < sysDim; sysCnt++)
                        {
                                modDim = (pModels[sysCnt])->dimension;

                                (pModels[sysCnt])->System(buf);

                                for(modCnt = 0; modCnt < modDim; modCnt++)
                                        ffunc[modCnt + shift] = buf[modCnt];

                                shift += modDim;
                        }
                }

                return ffunc;
        }

        static T& System(const std::valarray<model<T,EXCEPTION> *>& pModels, 
                T& time, std::valarray<T>& func) 
        { 
                size_t cnt, shift = GetDimension(pModels);
                
                MDS_ASSERT(func.size() >= shift, EXCEPTION,
                        "[model] insufficient size of valarray (argument 3) in System(const valarray<model<T,EXCEPTION> *>&, T&, valarray<T>&)");

                std::valarray<T> buf(shift + 1);
                
                System(pModels, buf, true);
                
                time = buf[0];

                for (cnt = 0; cnt < shift; cnt++)
                        func[cnt] = buf[cnt + 1];

                return time;
        }

        static std::valarray<T> System(const std::valarray<model<T,EXCEPTION> *>& pModels, bool isEnlargedPhase = false) 
        { 
                size_t shift = GetDimension(pModels);
                if (isEnlargedPhase) shift++;

                std::valarray<T> buf(shift);

                System(pModels, buf, isEnlargedPhase);
                
                return buf;
        }

        std::valarray<T>& DoSystem(std::valarray<T>& ffunc,
                const std::valarray<T>& arg, bool isEnlargedPhase = false)
        {                
                MDS_ASSERT((isEnlargedPhase && ffunc.size() >= dimension + 1) || (!isEnlargedPhase && ffunc.size() >= dimension), EXCEPTION,
                        "[model] insufficient size of valarray (argument 1) in DoSystem(valarray<T>&, const valarray<T>&, bool)");
                
                MDS_ASSERT(arg.size() == dimension + 1, EXCEPTION,
                        "[model] invalid size of valarray (argument 2) in DoSystem(valarray<T>&, const valarray<T>&, bool)");

                SetStatus(arg);

                return System(ffunc, isEnlargedPhase);
        }

        std::valarray<T> DoSystem(const std::valarray<T>& arg, bool isEnlargedPhase = false)
        {
                MDS_ASSERT(arg.size() == dimension + 1, EXCEPTION,
                        "[model] invalid size of valarray in DoSystem(const valarray<T>&, bool)");

                
                size_t dim = dimension;
                if (isEnlargedPhase) dim++;

                std::valarray<T> ffunc(dim);

                DoSystem(ffunc, arg, isEnlargedPhase);

                return ffunc;
        }

        std::valarray<T>& DoSystem(std::valarray<T>& ffunc, const T& arg_t,
                const std::valarray<T>& arg_x, bool isEnlargedPhase = false)
        {
                MDS_ASSERT((isEnlargedPhase && ffunc.size() >= dimension + 1) || (!isEnlargedPhase && ffunc.size() >= dimension), EXCEPTION,
                        "[model] insufficient size of valarray (argument 1) in DoSystem(valarray<T>&, const T&, const valarray<T>&, bool)");
                
                MDS_ASSERT(arg_x.size() == dimension + 1, EXCEPTION,
                        "[model] invalid size of valarray (argument 3) in DoSystem(valarray<T>&, const T&, const valarray<T>&, bool)");

                SetStatus(arg_t, arg_x);

                return System(ffunc, isEnlargedPhase);
        }

        std::valarray<T> DoSystem(const T& arg_t, const std::valarray<T>& arg_x, bool isEnlargedPhase = false)
        {
                MDS_ASSERT(arg_x.size() == dimension, EXCEPTION,
                        "[model] invalid size of valarray in DoSystem(const T&, const valarray<T>&, bool)");

                size_t dim = dimension;
                if (isEnlargedPhase) dim++;

                std::valarray<T> ffunc(dim);

                DoSystem(ffunc, arg_t, arg_x, isEnlargedPhase);
                
                return ffunc;
        }

        std::valarray<T>& DoSystem(std::valarray<T>& ffunc, 
                const T& arg_t, bool isEnlargedPhase = false)
        {
                MDS_ASSERT((isEnlargedPhase && ffunc.size() >= dimension + 1) || (!isEnlargedPhase && ffunc.size() >= dimension), EXCEPTION,
                        "[model] insufficient size of valarray in DoSystem(valarray<T>&, const T&, bool)");
                
                SetStatus(arg_t);

                return System(ffunc, isEnlargedPhase);
        }

        std::valarray<T> DoSystem(const T& arg_t, bool isEnlargedPhase = false)
        {
                size_t dim = dimension;
                if (isEnlargedPhase) dim++;

                std::valarray<T> ffunc(dim);

                DoSystem(ffunc, arg_t, isEnlargedPhase);
                
                return ffunc;
        }

        static std::valarray<T>& DoSystem(const std::valarray<model<T,EXCEPTION> *>& pModels, 
                std::valarray<T>& ffunc, const std::valarray<T>& arg, bool isEnlargedPhase = false)
        {
                MDS_ASSERT((isEnlargedPhase && ffunc.size() >= GetDimension(pModels) + 1) || (!isEnlargedPhase && ffunc.size() >= GetDimension(pModels)), EXCEPTION,
                        "[model] insufficient size of valarray (argument 2) in DoSystem(const valarray<model<T,EXCEPTION> *>&, valarray<T>&, const valarray<T>&, bool)");
                
                MDS_ASSERT(arg.size() == GetDimension(pModels) + 1, EXCEPTION,
                        "[model] invalid size of valarray (argument 3) in DoSystem(const valarray<model<T,EXCEPTION> *>&, valarray<T>&, const valarray<T>&, bool)");

                SetStatus(pModels, arg);

                return System(pModels, ffunc, isEnlargedPhase);
        }

        static std::valarray<T> DoSystem(const std::valarray<model<T,EXCEPTION> *>& pModels, 
                const std::valarray<T>& arg, bool isEnlargedPhase = false)
        {
                MDS_ASSERT(arg.size() == GetDimension(pModels) + 1, EXCEPTION,
                        "[model] invalid size of valarray (argument 2) in DoSystem(const valarray<model<T,EXCEPTION> *>&, const valarray<T>&, bool)");
                
                size_t dim = GetDimension(pModels);
                if (isEnlargedPhase) dim++;

                std::valarray<T> ffunc(dim);

                DoSystem(pModels, ffunc, arg, isEnlargedPhase);
                
                return ffunc;
        }

        static std::valarray<T>& DoSystem(const std::valarray<model<T,EXCEPTION> *>& pModels, 
                std::valarray<T>& ffunc, const T& arg_t, const std::valarray<T>& arg_x, bool isEnlargedPhase = false)
        {
                MDS_ASSERT((isEnlargedPhase && ffunc.size() >= GetDimension(pModels) + 1) || (!isEnlargedPhase && ffunc.size() >= GetDimension(pModels)), EXCEPTION,
                        "[model] insufficient size of valarray (argument 2) in DoSystem(const valarray<model<T,EXCEPTION> *>&, valarray<T>&, const T&, const valarray<T>&, bool)");

                MDS_ASSERT(arg_x.size() == GetDimension(pModels), EXCEPTION,
                        "[model] invalid size of valarray (argument 4) in DoSystem(const valarray<model<T,EXCEPTION> *>&, valarray<T>&, const T&, const valarray<T>&, bool)");

                SetStatus(pModels, arg_t, arg_x);

                return System(pModels, ffunc, isEnlargedPhase);
        }

        static std::valarray<T> DoSystem(const std::valarray<model<T,EXCEPTION> *>& pModels, 
                const T& arg_t, const std::valarray<T>& arg_x, bool isEnlargedPhase = false)
        {
                MDS_ASSERT(arg_x.size() == GetDimension(pModels), EXCEPTION,
                        "[model] invalid size of valarray (argument 3) in DoSystem(const valarray<model<T,EXCEPTION> *>&, const T&, const valarray<T>&, bool)");

                size_t dim = GetDimension(pModels);
                if (isEnlargedPhase) dim++;

                std::valarray<T> ffunc(dim);

                DoSystem(pModels, ffunc, arg_t, arg_x, isEnlargedPhase);
                
                return ffunc;
        }

        static std::valarray<T>& DoSystem(const std::valarray<model<T,EXCEPTION> *>& pModels, 
                std::valarray<T>& ffunc, const T& arg_t, bool isEnlargedPhase = false)
        {
                MDS_ASSERT((isEnlargedPhase && ffunc.size() >= GetDimension(pModels) + 1) || (!isEnlargedPhase && ffunc.size() >= GetDimension(pModels)), EXCEPTION,
                        "[model] insufficient size of valarray (argument 2) in DoSystem(const valarray<model<T,EXCEPTION> *>&, valarray<T>&, const T&, bool)");
                
                SetStatus(pModels, arg_t);

                return System(pModels, ffunc, isEnlargedPhase);
        }

        static std::valarray<T> DoSystem(const std::valarray<model<T,EXCEPTION> *>& pModels, 
                const T& arg_t, bool isEnlargedPhase = false)
        {
                size_t dim = GetDimension(pModels);
                if (isEnlargedPhase) dim++;

                std::valarray<T> ffunc(dim);
                
                DoSystem(pModels, ffunc, arg_t, isEnlargedPhase);
                
                return ffunc;
        }

// status & system status
// setting
        void SetStatus(const std::valarray<T>& arg) 
        { 
                MDS_ASSERT(arg.size() == dimension + 1, EXCEPTION,
                        "[model] invalid size of valarray in SetStatus(const valarray<T>&)");

                x = arg[phase]; t = arg[0];
        }

        void SetStatus(const T& arg_t, const std::valarray<T>& arg_x = std::valarray<T>(size_t(0)))
        { 
                MDS_ASSERT(!arg_x.size() || arg_x.size() == dimension, EXCEPTION,
                        "[model] invalid size of valarray in SetStatus(const T&, const valarray<T>&)");

                t = arg_t;

                if (arg_x.size()) x = arg_x;
        }

        static void SetStatus(const std::valarray<model<T,EXCEPTION> *>& pModels, const std::valarray<T>& arg) 
        { 
                size_t sysCnt, modCnt, shift=0, modDim, sysDim = pModels.size();
                
                MDS_ASSERT(arg.size() == GetDimension(pModels) + 1, EXCEPTION,
                        "[model] invalid size of valarray (argument 2) in SetStatus(const valarray<model<T,EXCEPTION> *>&, const valarray<T>&)");

                for (sysCnt = 0; sysCnt < sysDim; sysCnt++)
                {
                        modDim = (pModels[sysCnt])->dimension;

                        for (modCnt = 0; modCnt < modDim; modCnt++)
                                (pModels[sysCnt])->x[modCnt] = arg[modCnt + shift + 1];

                        (pModels[sysCnt])->t = arg[0];
                        
                        shift += modDim;
                }
        }

        static void SetStatus(const std::valarray<model<T,EXCEPTION> *>& pModels, 
                const T &arg_t, const std::valarray<T>& arg_x = std::valarray<T>(size_t(0))) 
        { 
                size_t sysCnt, modCnt, shift=0, modDim, sysDim = pModels.size();
                
                MDS_ASSERT(!arg_x.size() || arg_x.size() == GetDimension(pModels), EXCEPTION,
                        "[model] invalid size of valarray (argument 3) in SetStatus(const valarray<model<T,EXCEPTION> *>&, const T&, const valarray<T>&)");
                                
                for (sysCnt = 0; sysCnt < sysDim; sysCnt++)
                {
                        if (arg_x.size())
                        {
                                modDim = (pModels[sysCnt])->dimension;

                                for (modCnt = 0; modCnt < modDim; modCnt++)
                                        (pModels[sysCnt])->x[modCnt] = arg_x[modCnt + shift];
                        }

                        (pModels[sysCnt])->t = arg_t;

                        shift += modDim;
                }
        }

// getting
        std::valarray<T>& GetStatus(std::valarray<T>& arg, bool isEnlargedPhase = false) const 
        { 
                MDS_ASSERT((isEnlargedPhase && arg.size() >= dimension + 1) || (!isEnlargedPhase && arg.size() >= dimension), EXCEPTION,
                        "[model] insufficient size of valarray in GetStatus(valarray<T>&, bool)");

                size_t cnt;

                if (isEnlargedPhase) 
                {
                        for (cnt = 0; cnt < dimension; cnt++) arg[cnt + 1] = x[cnt];
                        
                        arg[0] = t;
                }
                else
                        for (cnt = 0; cnt < dimension; cnt++) arg[cnt] = x[cnt];

                return arg;
        }

        T& GetStatus(T& arg_t, std::valarray<T>& arg_x) const 
        { 
                MDS_ASSERT(arg_x.size() >= dimension, EXCEPTION,
                        "[model] insufficient size of valarray in GetStatus(T&, valarray<T>&)");

                size_t cnt;

                arg_t = t;

                for (cnt = 0; cnt < dimension; cnt++)
                        arg_x[cnt] = x[cnt];

                return arg_t;
        }

        T& GetStatus(T& arg_t) const 
        { 
                arg_t = t;

                return arg_t;
        }

        std::valarray<T> GetStatus(bool isEnlargedPhase = false) const
        { 
                if (isEnlargedPhase) 
                {
                        std::valarray<T> status(dimension+1);

                        status[phase] = x; status[0] = t;

                        return status;
                }

                return x;
        }

        static std::valarray<T>& GetStatus(const std::valarray<model<T,EXCEPTION> *>& pModels, 
                std::valarray<T>& arg, bool isEnlargedPhase = false) 
        { 
                size_t sysCnt, modCnt, shift = 0, modDim, sysDim = pModels.size();
                
                MDS_ASSERT(pModels.size(), EXCEPTION,
                        "[model] empty valarray of models (argument 1) in GetStatus(const valarray<model<T,EXCEPTION> *>&, valarray<T>&, bool)");

                MDS_ASSERT((isEnlargedPhase && arg.size() >= GetDimension(pModels) + 1) || (!isEnlargedPhase && arg.size() >= GetDimension(pModels)), EXCEPTION,
                        "[model] insufficient size of valarray (argument 2) in GetStatus(const valarray<model<T,EXCEPTION> *>&, valarray<T>&, bool)");

                if (isEnlargedPhase) 
                {
                        for (sysCnt = 0; sysCnt < sysDim; sysCnt++)
                        {
                                modDim = (pModels[sysCnt])->dimension;

                                for(modCnt = 0; modCnt < modDim; modCnt++)
                                        arg[modCnt + shift + 1] = (pModels[sysCnt])->x[modCnt];

                                shift += modDim;
                        }
                        
                        arg[0] = (pModels[0])->t;
                }
                else
                {
                        for (sysCnt = 0; sysCnt < sysDim; sysCnt++)
                        {
                                modDim = (pModels[sysCnt])->dimension;

                                for (modCnt = 0; modCnt < modDim; modCnt++)
                                        arg[modCnt + shift] = (pModels[sysCnt])->x[modCnt];

                                shift += modDim;
                        }                        
                }

                return arg;
        }

        static T& GetStatus(const std::valarray<model<T,EXCEPTION> *>& pModels, 
                T &arg_t, std::valarray<T> &arg_x) 
        { 
                size_t cnt, shift = GetDimension(pModels);

                std::valarray<T> arg(shift + 1);

                MDS_ASSERT(pModels.size(), EXCEPTION,
                        "[model] empty valarray of models (argument 1) in GetStatus(const valarray<model<T,EXCEPTION> *>&, T&, valarray<T>&)");

                MDS_ASSERT(arg_x.size() >= shift, EXCEPTION,
                        "[model] insufficient size of valarray (argument 3) in GetStatus(const valarray<model<T,EXCEPTION> *>&, T&, valarray<T>&)");

                GetStatus(pModels, arg, true);

                arg_t = arg[0];

                for (cnt = 0; cnt < shift; cnt++)
                        arg_x[cnt] = arg[cnt + 1];

                return arg_t;
        }

        static T& GetStatus(const std::valarray<model<T,EXCEPTION> *>& pModels, T& arg_t) 
        { 
                MDS_ASSERT(pModels.size(), EXCEPTION,
                        "[model] empty valarray of models in GetStatus(const valarray<model<T,EXCEPTION> *>&, T&)");

                arg_t = (pModels[0])->t;

                return arg_t;
        }

        static std::valarray<T> GetStatus(const std::valarray<model<T,EXCEPTION> *>& pModels, 
                bool isEnlargedPhase = false) 
        { 
                MDS_ASSERT(pModels.size(), EXCEPTION,
                        "[model] empty valarray of models in GetStatus(const valarray<model<T,EXCEPTION> *>&, bool)");

                size_t shift = GetDimension(pModels);
                if (isEnlargedPhase) shift++;

                std::valarray<T> arg(shift);

                GetStatus(pModels, arg, isEnlargedPhase);
                
                return arg;
        }

        std::valarray<T>& GetSystemStatus(std::valarray<T>& arg, bool isEnlargedPhase = false) const 
        { 
                MDS_ASSERT((isEnlargedPhase && arg.size() >= dimension + 1) || (!isEnlargedPhase && arg.size() >= dimension), EXCEPTION,
                        "[model] insufficient size of valarray in GetSystemStatus(valarray<T>&, bool)");

                size_t cnt;

                if (isEnlargedPhase) 
                {
                        for (cnt = 0; cnt < dimension; cnt++) arg[cnt + 1] = y[cnt];

                        arg[0] = t;
                }
                else
                        for (cnt = 0; cnt < dimension; cnt++) arg[cnt] = y[cnt];

                return arg;
        }

        T& GetSystemStatus(T& arg_t, std::valarray<T>& arg_x) const 
        { 
                MDS_ASSERT(arg_x.size() >= dimension, EXCEPTION,
                        "[model] insufficient size of valarray in GetSystemStatus(T&, valarray<T>&)");

                size_t cnt;

                arg_t = t;

                for (cnt = 0; cnt < dimension; cnt++)
                        arg_x[cnt] = y[cnt];

                return arg_t;
        }

        std::valarray<T> GetSystemStatus(bool isEnlargedPhase = false) const
        { 
                if (isEnlargedPhase)
                {
                        std::valarray<T> status(dimension + 1);

                        status[phase] = y; status[0] = t;

                        return status;
                }

                return y;
        }

        static std::valarray<T>& GetSystemStatus(const std::valarray<model<T,EXCEPTION> *>& pModels, 
                std::valarray<T>& arg, bool isEnlargedPhase = false) 
        { 
                size_t sysCnt, modCnt, modDim, shift = 0, sysDim = pModels.size();
                
                MDS_ASSERT(pModels.size(), EXCEPTION,
                        "[model] empty valarray of models in GetSystemStatus(const valarray<model<T,EXCEPTION> *>&, valarray<T>&, bool)");

                MDS_ASSERT((isEnlargedPhase && arg.size() >= GetDimension(pModels) + 1) || (!isEnlargedPhase && arg.size() >= GetDimension(pModels)), EXCEPTION,
                        "[model] insufficient size of valarray (argument 2) in GetSystemStatus(const valarray<model<T,EXCEPTION> *>&, valarray<T>&, bool)");

                if (isEnlargedPhase) 
                {
                        for (sysCnt = 0; sysCnt < sysDim; sysCnt++)
                        {
                                modDim = (pModels[sysCnt])->dimension;

                                for (modCnt = 0; modCnt < modDim; modCnt++)
                                        arg[modCnt + shift + 1] = (pModels[sysCnt])->y[modCnt];
                                
                                shift += modDim;
                        }
                        
                        arg[0] = (pModels[0])->t;
                }
                else
                {
                        for (sysCnt = 0; sysCnt < sysDim; sysCnt++)
                        {
                                modDim = (pModels[sysCnt])->dimension;

                                for (modCnt = 0; modCnt < modDim; modCnt++)
                                        arg[modCnt + shift] = (pModels[sysCnt])->y[modCnt];

                                shift += modDim;
                        }                        
                }

                return arg;
        }

        static T& GetSystemStatus(const std::valarray<model<T,EXCEPTION> *>& pModels, 
                T& arg_t, std::valarray<T>& arg_x) 
        { 
                size_t cnt, shift = GetDimension(pModels);

                std::valarray<T> arg(shift + 1);

                MDS_ASSERT(pModels.size(), EXCEPTION,
                        "[model] empty valarray of models in GetSystemStatus(const valarray<model<T,EXCEPTION> *>&, T&, valarray<T>&)");

                MDS_ASSERT(arg_x.size() >= shift, EXCEPTION,
                        "[model] insufficient size of valarray (argument 3) in GetSystemStatus(const valarray<model<T,EXCEPTION> *>&, T&, valarray<T>&)");

                GetSystemStatus(pModels, arg, true);

                arg_t = arg[0];

                for (cnt = 0; cnt < shift; cnt++)
                        arg_x[cnt] = arg[cnt + 1];

                return arg_t;
        }

        static std::valarray<T> GetSystemStatus(const std::valarray<model<T,EXCEPTION> *>& pModels, 
                bool isEnlargedPhase = false) 
        { 
                MDS_ASSERT(pModels.size(), EXCEPTION,
                        "[model] empty valarray of models in GetSystemStatus(const valarray<model<T,EXCEPTION> *>&, bool)");

                size_t shift = GetDimension(pModels);
                if (isEnlargedPhase) shift++;

                std::valarray<T> arg(shift);

                GetSystemStatus(pModels, arg, isEnlargedPhase);
                
                return arg;
        }

//model time manipulation
        T& GetTime(T& arg) const
        { 
                arg = t;

                return arg;
        }

        T GetTime() const
        { 
                return t;
        }

        T& GetTime(const std::valarray<model<T,EXCEPTION> *>& pModels, T& arg) const
        { 
                MDS_ASSERT(pModels.size(), EXCEPTION,
                        "[model] empty valarray of models in GetTime(const valarray<model<T,EXCEPTION> *>&, T&)");

                arg = (pModels[0])->t;

                return arg;
        }

        T GetTime(const std::valarray<model<T,EXCEPTION> *>& pModels) const
        { 
                MDS_ASSERT(pModels.size(), EXCEPTION,
                        "[model] empty valarray of models in GetTime(const valarray<model<T,EXCEPTION> *>&)");

                return (pModels[0])->t;
        }

        void SetTime(const T &arg)
        { 
                t = arg;
        }

        void SetTime(const std::valarray<model<T,EXCEPTION> *>& pModels, const T& arg) 
        { 
                size_t sysCnt, sysDim = pModels.size();

                for (sysCnt = 0; sysCnt < sysDim; sysCnt++)
                        (pModels[sysCnt])->t = arg;
        }

// id, dimension, emulation
        size_t emulation_order;  // order of approximation in emulation mode

        result_manager<T,T,EXCEPTION> emulation_table;  // for record emulation chain

        size_t GetDimension() const 
        { 
                return dimension; 
        } 

        static size_t GetDimension(const std::valarray<model<T,EXCEPTION> *>& pModels) 
        { 
                size_t sysCnt, shift = 0, sysDim = pModels.size();

                for (sysCnt = 0; sysCnt < sysDim; sysCnt++)
                        shift += (pModels[sysCnt])->dimension;
                
                return shift; 
        }

        int GetIdentify() const 
        { 
                return identify; 
        } 

        int GetEmulationMode() const 
        { 
                return emulation_mode; 
        }
        
        void SetEmulationMode(int newMode) 
        { 
                switch (newMode)
                {
                case EMULATION_ENABLED: 
                case EMULATION_RECORDING: emulation_mode = newMode; break; 

                default: emulation_mode = EMULATION_DISABLED; 
                }
        }

// return phaze: mask for phase vector
        std::valarray<bool> PhaseMask() const
        { 
                return phase; 
        }

        
protected:
        T t;                // current time
        std::valarray<T> x; // current phase
        std::valarray<T> y; // current system value

        int identify;       // system identifying

private:
        std::valarray<bool> phase; // mask for obtaining phase from status
        size_t dimension;          // system dimension (quantity of equations)
        int emulation_mode;        // emulation disabled(0), enabled(1), record(2)
};


#endif  
