//***************************************************************************
//
//   class cauchy_region (from MDS v.5.2)
//   encapsulates start condition (a.k.a. Cauchy condition) & final condition
//   (c) 2002-2004 Dr.Bashnyakov A. 
//   (bash@unicyb.kiev.ua)
//
//***************************************************************************
//                                                                         
//   This program is free software; you can redistribute it and/or modify  
//   it under the terms of the GNU General Public License as published by  
//   the Free Software Foundation; either version 2 of the License, or     
//   (at your option) any later version.                                   
//                                                                         
//***************************************************************************


#ifndef _region_h_
#define _region_h_


#include <valarray>
#include "mds_error.h"
#include "mds_assert.h"


template <class T = MDS_DEFAULT_TYPE, class EXCEPTION = MDS_DEFAULT_EXCEPTION>
class cauchy_region
{
private:
// Cauchy condition x(t0)=x0 or t0 only
        T t0;
        std::valarray<T> x0;

// Final condition: 
// WARNING: isRegion() use Region() if Region not null  _OR_  tFinal
// simple final condition: time of end of algorithm. 
        T tFinal;

// must return: 1-IN region, -1-OUT region, 0-at region bound        
        int(*Region)(const T&, const std::valarray<T>&);

        std::valarray<bool> mask;  // mask for phase separating

public:
// base constructor: Cauchy condition & region function
        cauchy_region(const std::valarray<T>& cauchyCond, int(*f)(const T&, const std::valarray<T>&)) :
                x0(cauchyCond.size() - 1), mask(true, cauchyCond.size()), Region(f) 
        {
                MDS_ASSERT(cauchyCond.size() >= 1, EXCEPTION,
                        "[cauchy_region] invalid size of valarray in constructor");
                
                mask[0] = false;

                t0 = cauchyCond[0]; x0 = cauchyCond[mask]; 
        }

// alt constructor: Cauchy condition & t final
        cauchy_region(const std::valarray<T>& cauchyCond, const T& t) :
                x0(cauchyCond.size() - 1), mask(true, cauchyCond.size()),
                Region(0), tFinal(t)
        { 
                MDS_ASSERT(cauchyCond.size() >= 1, EXCEPTION,
                        "[cauchy_region] invalid size of valarray in constructor");
                
                mask[0] = false;

                t0 = cauchyCond[0]; x0 = cauchyCond[mask]; 
        }

// alt constructor: Cauchy condition & region function
        cauchy_region(const T &start, const std::valarray<T>& cauchyCond, int(*f)(const T&, const std::valarray<T>&)) :
                t0(start), x0(cauchyCond), mask(true, cauchyCond.size() + 1), Region(f) 
        {
                mask[0] = false;
        }

// alt constructor: Cauchy condition & t final
        cauchy_region(const T &start, const std::valarray<T>& cauchyCond, const T& t) :
                t0(start), x0(cauchyCond), mask(true, cauchyCond.size() + 1), 
                Region(0), tFinal(t)
        { 
                mask[0] = false;
        }

// alt constructor: t start & region function
        cauchy_region(const T& start, int(*f)(const T&, const std::valarray<T>&)) :
                t0(start), Region(f), mask(false, 1)
        {}

// simplest constructor: t start & t final
        cauchy_region(const T& start, const T& fin) :
                t0(start), Region(0), tFinal(fin), mask(false, 1)
        {}

// base region method 
// return: 
//   1-status is IN region, 
//  -1-status is OUT region,
//   0-status is at region bound
        int isRegion(const T& t, const std::valarray<T>& x = std::valarray<T>(size_t(0))) const
        {
                MDS_ASSERT(!x.size() || !x0.size() || x.size() == x0.size(), EXCEPTION,
                        "[cauchy_region] invalid size of valarray in isRegion(const T&, const valarray<T>&)");

                if (Region) return Region(t, x);
                else 
                {
                        if ( (t > t0 && t < tFinal) || (t > tFinal && t < t0) ) return 1;
                        else if (t == t0 || t == tFinal) return 0;

                        return -1;
                }
        }

        int isRegion(const std::valarray<T>& x) const
        {
                MDS_ASSERT(!x.size() || !x0.size() || x.size() == x0.size() + 1, EXCEPTION,
                        "[cauchy_region] invalid size of valarray in isRegion(const valarray<T>&)");

                return isRegion(x[0], x[mask]);
        }

// getting parameters
        T& GetStartTime(T& arg) const
        { 
                arg = t0;

                return arg;
        }

        T GetStartTime() const
        { 
                return t0;
        }

        T& GetFinalTime(T& arg) const
        { 
                arg = tFinal;

                return arg;
        }

        T GetFinalTime() const
        { 
                return tFinal;
        }

        std::valarray<T>& GetCauchyCondition(std::valarray<T>& arg, bool isEnlargedPhase = false) const 
        { 
                size_t dim = x0.size();

                MDS_ASSERT((isEnlargedPhase && arg.size() >= dim + 1) || (!isEnlargedPhase && arg.size() >= dim), EXCEPTION,
                        "[cauchy_region] insufficient size of valarray in GetCauchyCondition(valarray<T>&, bool)");

                size_t cnt;

                if (isEnlargedPhase) 
                {
                        for (cnt = 0; cnt < dim; cnt++) arg[cnt + 1] = x0[cnt];

                        arg[0] = t0;
                }
                else
                        for (cnt = 0; cnt < dim; cnt++) arg[cnt] = x0[cnt];

                return arg;
        }

        T& GetCauchyCondition(T& arg_t, std::valarray<T>& arg_x) const 
        { 
                size_t dim = x0.size();

                MDS_ASSERT(arg_x.size() >= dim, EXCEPTION,
                        "[cauchy_region] insufficient size of valarray in GetCauchyCondition(T&, valarray<T>&)");
                
                size_t cnt;
                
                arg_t = t0;

                for (cnt = 0; cnt < dim; cnt++)
                        arg_x[cnt] = x0[cnt];

                return arg_t;
        }

        T& GetCauchyCondition(T& arg_t) const 
        { 
                arg_t = t0;

                return arg_t;
        }

        std::valarray<T> GetCauchyCondition(bool isEnlargedPhase = false) const
        { 
                if (isEnlargedPhase) 
                {
                        std::valarray<T> status(x0.size() + 1);

                        status[mask] = x0; status[0] = t0;

                        return status;
                }

                return x0;
        }

        size_t GetCauchyConditionSize() const
        { 
                return x0.size();
        }

// what is the final condition, is it function? 
// return: 0-no (use t final), 1-yes (use Region())
        int isFuncFinalCondition() const
        { 
                if (Region) return 1;
                return 0;
        }

};


#endif  
