//***************************************************************************
//
//   class result_manager (from MDS v.5.2)
//   accumulation and saving results of calculations  
//   (c) 2002-2004 Dr.Bashnyakov A. 
//   (bash@unicyb.kiev.ua)
//
//***************************************************************************
//                                                                         
//   This program is free software; you can redistribute it and/or modify  
//   it under the terms of the GNU General Public License as published by  
//   the Free Software Foundation; either version 2 of the License, or     
//   (at your option) any later version.                                   
//                                                                         
//***************************************************************************


#ifndef _result_h_
#define _result_h_


#include <map>
#include <valarray>
#include <utility>
#include <string>
#include <fstream>
#include <sstream>
#include "mds_error.h"
#include "mds_assert.h"


// class for menage the calculation results
template <class Tind = MDS_DEFAULT_TYPE, class Tdata = Tind, class EXCEPTION = MDS_DEFAULT_EXCEPTION, class Cmp = std::less<Tind>, class A = std::allocator<std::pair<const Tind, std::valarray<Tdata> > > >
class result_manager : public std::multimap<Tind, std::valarray<Tdata>, Cmp, A> 
{
private:
        size_t cntDensity;  // temp var for adding data
        std::string autosave_comment;    // comment in file saved by destructor
        std::string autosave_file_name;  // name of file saved by destructor

public:
        bool autosave;   // if true - destructor will save this 
        size_t density;  // adding density
        char separator, remark_mark;  // mark between data & mark of remark 
        
        explicit result_manager(const char* name = "", const char* comm = "")
                : density(1), autosave(true), separator(' '), remark_mark(';'),
                autosave_comment(comm), autosave_file_name(name), cntDensity(1)
        {}

        virtual ~result_manager()
        {
                if (autosave) Save(autosave_file_name.c_str(), autosave_comment.c_str(), 1);
        }

// adding element to container
// return: 0-OK, 1-density violation
// add src(arg), src dimension is n 
        int Add(const Tind& arg, const std::valarray<Tdata>& src, bool isReplace = false)
        {
                MDS_ASSERT(!size() || (begin()->second).size() == src.size(), EXCEPTION,
                        "[result_manager] invalid size of valarray in Add(const Tind&, const valarray<Tdata>&, bool)");

                if (!density) return 1;

                if (cntDensity >= density)
                {
                        if (isReplace) erase(arg);

                        insert(std::pair<Tind, std::valarray<Tdata> > (arg, src));
                        
                        cntDensity = 1;
                        
                        return 0;
                }

                cntDensity++;
                
                return 1;
        }

// add src(arg), src dimension is 0, e.g. we are the set of arguments 
        int Add(const Tind& arg, bool isReplace = false)
        {
                return Add(arg, std::valarray<Tdata>(size_t(0)));
        }

// add src(arg), src dimension is 1 
        int Add(const Tind& arg, const Tdata& src, bool isReplace = false)
        {
                return Add(arg, std::valarray<Tdata>(src, 1));
        }
        
// add another result_manager 
        int Add(const result_manager<Tind, Tdata, EXCEPTION, Cmp, A>& src, bool isReplace = false)
        {
                MDS_ASSERT(!size() || !src.size() || (begin()->second).size() == (src.begin()->second).size(), EXCEPTION,
                        "[result_manager] invalid size of valarray in Add(const result_manager&, bool)");

                int return_value = 1;

                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::const_iterator mapIterator = src.begin();
                
                while (mapIterator != src.end())
                {
                        if (!Add(mapIterator->first, mapIterator->second, isReplace))
                                return_value = 0;
                        
                        mapIterator++;
                }
                
                return return_value; 
        }
   
// simple data manipulation
        result_manager<Tind, Tdata, EXCEPTION, Cmp, A>& operator += (const Tdata& src)
        {
                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::iterator mapIterator = begin();
                
                while (mapIterator != end())
                {
                        mapIterator->second += src;
                
                        mapIterator++;
                }

                return *this; 
        }

        result_manager<Tind, Tdata, EXCEPTION, Cmp, A>& operator -= (const Tdata& src)
        {
                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::iterator mapIterator = begin();
                
                while (mapIterator != end())
                {
                        mapIterator->second -= src;

                        mapIterator++;
                }

                return *this; 
        }

        result_manager<Tind, Tdata, EXCEPTION, Cmp, A>& operator *= (const Tdata& src)
        {
                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::iterator mapIterator = begin();

                while (mapIterator != end())
                {
                        mapIterator->second *= src;

                        mapIterator++;
                }

                return *this; 
        }

        result_manager<Tind, Tdata, EXCEPTION, Cmp, A>& operator /= (const Tdata& src)
        {
                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::iterator mapIterator = begin();

                while (mapIterator != end())
                {
                        mapIterator->second /= src;

                        mapIterator++;
                }

                return *this; 
        }

// saving container to file
// return: 0-OK, 1-density violation, 2-cannot open file for saving 
        int Save(const char* fName, 
                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::const_iterator first,
                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::const_iterator last,
                const std::valarray<bool>& mask = std::valarray<bool>(size_t(0)),
                const char* comm = "", size_t dens = 1) const
        {
                MDS_ASSERT(!mask.size() || (first->second).size() + 1 >= mask.size(), EXCEPTION,
                        "[result_manager] invalid size of mask in Save(const char*, const_iterator, const_iterator, const valarray<bool>&, const char*, size_t)");

                if (!dens) return 1;
                
                size_t cnt, cntDens = dens, phazeSize;

                bool isMask (mask.size() != 0);
                if (isMask) phazeSize = mask.size() - 1;
                else if(first != end()) phazeSize = (first->second).size();

                if (!fName || !strlen(fName)) return 2;
                std::ofstream stream(fName);
                if (!stream) return 2;

                std::string temp(comm);

                if (temp.length()) 
                {
                        stream << remark_mark;

                        for (cnt = 0; cnt < temp.length(); cnt++) 
                        {
                                if (temp[cnt] == '\n') stream << std::endl << remark_mark;
                                else stream << temp[cnt];
                        }

                        stream << std::endl << std::endl;
                }
                
                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::const_iterator mapIterator = first;

                while (mapIterator != last)
                {        
                        if (dens == cntDens)
                        {
                                if (!isMask || mask[0]) 
                                {
                                        stream << mapIterator->first;

                                        if (separator != ' ') stream << separator;

                                        stream << ' ';
                                }

                                for (cnt = 0; cnt < phazeSize; cnt++)
                                        if (!isMask || mask[cnt + 1])
                                        {
                                                stream << (mapIterator->second)[cnt];

                                                if (separator != ' ' && cnt < phazeSize - 1) 
                                                        stream << separator;
                                                
                                                if (cnt < phazeSize - 1) stream << ' ';
                                        }

                                stream << std::endl;

                                cntDens = 1;        
                        }
                        else cntDens++;

                        mapIterator++;
                }
                
                stream << std::flush;
                
                return 0;
        }

        int Save(const char* fName, typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::const_iterator first,
                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::const_iterator last, const char* comm = "", size_t dens = 1)
        {
                return Save(fName, first, last, std::valarray<bool>(size_t(0)), comm, dens);
        }

        int Save(const char* fName, const std::valarray<bool>& mask, const char* comm = "", size_t dens = 1) const
        {
                MDS_ASSERT(begin() == end() || !mask.size() || (begin()->second).size() + 1 >= mask.size(), EXCEPTION,
                        "[result_manager] invalid size of mask in Save(const char*, const valarray<bool>&, const char*, size_t)");

                return Save(fName, begin(), end(), mask, comm, dens);
        }

        int Save(const char* fName, const char* comm = "", size_t dens = 1) const
        {
                return Save(fName, begin(), end(), std::valarray<bool>(size_t(0)), comm, dens);
        }

// loading container from file
// return: 0-OK, 1-density violation, 2-cannot open file  
        int Load(const char* fName, bool isReplace = false)
        {
                if (!fName || !strlen(fName)) return 2;

                std::ifstream stream(fName);
                if (!stream) return 2;

                std::string str;
                
                size_t dim = 0, i, j;
                
                bool isFind = false, isEmptyStr, isLastSpace = true;

                while (!isFind && std::getline(stream, str))
                {
                        i = 0; j = str.length();

                        while (i < j && isspace(str[i])) i++;
                        
                        if (i < j && str[i] != remark_mark) isFind = true;
                }
                
                if (isFind) 
                {        
                        for (i = 0; i < j; i++)
                                if (str[i] == separator || isspace(str[i]))
                                {
                                        if (!isLastSpace)
                                        {
                                                dim++;

                                                isLastSpace = true;
                                        }
                                }
                                else isLastSpace = false;
                }
                else return 1;

                MDS_ASSERT(!size() || dim == (begin()->second).size(), EXCEPTION,
                        "[result_manager] incorrect loading by Load(const char*, bool)");

                std::valarray<Tdata> arg_x(dim);

                Tind arg_t;
                
                int return_value = 1;

                do
                {
                        isEmptyStr = true;

                        j = str.length();

                        for (i = 0; i < j; i++)
                        {
                                if (str[i] == separator) str[i] = ' ';

                                if (str[i] != ' ') isEmptyStr = false;
                        }
                        
                        if (isEmptyStr) continue;

                        std::istringstream strstr(str);

                        strstr >> arg_t;
                        
                        for (i = 0; i < dim; i++) strstr >> arg_x[i];

                        if (!Add(arg_t, arg_x, isReplace)) return_value = 0;

                } while (std::getline(stream, str));
                
                return return_value;
        }

// apply function for all element in container within mask
        void Modify(typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::iterator first,
                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::iterator last,
                Tdata fn(Tdata), const std::valarray<bool>& mask = std::valarray<bool>(size_t(0)))
        {
                MDS_ASSERT(!mask.size() || (first->second).size() >= mask.size(), EXCEPTION,
                        "[result_manager] invalid size of mask in Modify(iterator, iterator, fn, const valarray<bool>&)");

                size_t cnt, phazeSize;

                bool isMask(mask.size() != 0);
                if (isMask) phazeSize = mask.size();
                else if(first != end()) phazeSize = (first->second).size();
                
                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::iterator mapIterator = first;
                
                while (mapIterator != last)
                {        
                        for (cnt = 0; cnt < phazeSize; cnt++)
                                if (!isMask || mask[cnt]) 
                                        (mapIterator->second)[cnt] = fn((mapIterator->second)[cnt]);

                        mapIterator++;
                }
        }
        
        void Modify(Tdata fn(Tdata), const std::valarray<bool>& mask = std::valarray<bool>(size_t(0)))
        {
                MDS_ASSERT(begin() == end() || !mask.size() || (begin()->second).size() >= mask.size(), EXCEPTION,
                        "[result_manager] invalid size of mask in Modify(fn, const valarray<bool>&)");

                Modify(begin(), end(), fn, mask);
        }

        void Modify(typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::iterator first,
                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::iterator last,
                Tdata fn(const Tdata&), const std::valarray<bool>& mask = std::valarray<bool>(size_t(0)))
        {
                MDS_ASSERT(!mask.size() || (first->second).size() >= mask.size(), EXCEPTION,
                        "[result_manager] invalid size of mask in Modify(iterator, iterator, fn, const valarray<bool>&)");

                size_t cnt, phazeSize;

                bool isMask(mask.size() != 0);
                if (isMask) phazeSize = mask.size();
                else if(first != end()) phazeSize = (first->second).size();
                
                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::iterator mapIterator = first;

                while (mapIterator != last)
                {        
                        for (cnt = 0; cnt < phazeSize; cnt++)
                                if (!isMask || mask[cnt]) 
                                        (mapIterator->second)[cnt] = fn((mapIterator->second)[cnt]);

                        mapIterator++;
                }
        }
        
        void Modify(Tdata fn(const Tdata&), const std::valarray<bool>& mask = std::valarray<bool>(size_t(0)))
        {
                MDS_ASSERT(begin() == end() || !mask.size() || (begin()->second).size() >= mask.size(), EXCEPTION,
                        "[result_manager] invalid size of mask in Modify(fn, const valarray<bool>&)");

                Modify(begin(), end(), fn, mask);
        }
        
// getting element from container
// return: quantity of data found with the key "t"
        size_t Get(std::valarray<Tdata>& x, const Tind& t) const
        {
                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::const_iterator exactFind = find(t);

                if (exactFind != end())
                {        
                        size_t cnt, dim = (exactFind->second).size();

                        MDS_ASSERT(x.size() == dim, EXCEPTION,
                                "[result_manager] insufficient size of valarray in Get(valarray<Tdata>&, const Tind&)");

                        for (cnt = 0; cnt < dim; cnt++)
                                x[cnt] = (exactFind->second)[cnt];
                }
                
                return count(t);
        }

// return data by argument t
        std::valarray<Tdata> Get(const Tind& t) const
        {
                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::const_iterator exactFind = find(t);

                MDS_ASSERT(exactFind != end(), EXCEPTION,
                        "[result_manager] no valarray returned by Get(const Tind&)");

                return exactFind->second;
        }

// bang-bang approximation
// return: divergence = t_closest - t_restricted
        Tind GetBangBang(std::valarray<Tdata>& x, const Tind& t, bool isNearest = false) const
        {
                MDS_ASSERT(size(), EXCEPTION,
                        "[result_manager] empty container in GetBangBang([valarray<Tdata>&,] const Tind&, bool)");

                MDS_ASSERT(x.size() >= (begin()->second).size(), EXCEPTION,
                        "[result_manager] insufficient size of valarray in GetBangBang([valarray<Tdata>&,] const Tind&, bool)");
                
                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::const_iterator exactFind = find(t);

                Tind returnValue;

                size_t cnt, dim = (begin()->second).size(); 
                
                if (exactFind != end())
                {
                        MDS_ASSERT(count(t) == 1, EXCEPTION,
                                "[result_manager] more than one result in GetBangBang([valarray<Tdata>&,] const Tind&, bool)");
                                
                        for (cnt = 0; cnt < dim; cnt++)
                                x[cnt] = (exactFind->second)[cnt];

                        returnValue = t - t;
                }
                else
                {
                        typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::const_iterator lowerKey = lower_bound(t);
                        typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::const_iterator upperKey = lowerKey;

                        if (upperKey == begin())
                        {
                                MDS_ASSERT(count(upperKey->first) == 1, EXCEPTION,
                                        "[result_manager] more than one result in GetBangBang([valarray<Tdata>&,] const Tind&, bool)");
                                        
                                for (cnt = 0; cnt < dim; cnt++)
                                        x[cnt] = (begin()->second)[cnt];

                                returnValue = begin()->first - t;
                        }
                        else
                        {
                                lowerKey--;

                                if (upperKey == end())
                                {
                                        MDS_ASSERT(count(lowerKey->first) == 1, EXCEPTION,
                                                "[result_manager] more than one result in GetBangBang([valarray<Tdata>&,] const Tind&, bool)");
                                        
                                        for (cnt = 0; cnt < dim; cnt++)
                                                x[cnt] = (lowerKey->second)[cnt];

                                        returnValue = lowerKey->first - t;
                                }
                                else
                                {
                                        if (isNearest)
                                        {
                                                if (t - lowerKey->first < upperKey->first - t) 
                                                {
                                                        MDS_ASSERT(count(lowerKey->first) == 1, EXCEPTION,
                                                                "[result_manager] more than one result in GetBangBang([valarray<Tdata>&,] const Tind&, bool)");
                                                                
                                                        for (cnt = 0; cnt < dim; cnt++)
                                                                x[cnt] = (lowerKey->second)[cnt];

                                                        returnValue = t - lowerKey->first;
                                                }
                                                else 
                                                {
                                                        MDS_ASSERT(count(upperKey->first) == 1, EXCEPTION,
                                                                "[result_manager] more than one result in GetBangBang([valarray<Tdata>&,] const Tind&, bool)");
                                        
                                                        for (cnt = 0; cnt < dim; cnt++)
                                                                x[cnt] = (upperKey->second)[cnt];

                                                        returnValue = upperKey->first - t;
                                                }
                                        }
                                        else
                                        {
                                                MDS_ASSERT(count(lowerKey->first) == 1, EXCEPTION,
                                                        "[result_manager] more than one result in GetBangBang([valarray<Tdata>&,] const Tind&, bool)");
                                        
                                                for (cnt = 0; cnt < dim; cnt++)
                                                        x[cnt] = (lowerKey->second)[cnt];

                                                returnValue = t - lowerKey->first;
                                        }
                                }
                        }
                }
                
                return returnValue;
        }

// bang-bang approximation
        std::valarray<Tdata> GetBangBang(const Tind& t, bool isNearest = false) const
        {
                MDS_ASSERT(size(), EXCEPTION,
                        "[result_manager] empty container in GetBangBang(const Tind&, bool)");

                std::valarray<Tdata> v((begin()->second).size());

                GetBangBang(v, t, isNearest);

                return v;
        }

// approximation with order (0:bang-bang,1:linear)
// return: divergence = t_closest - t_restricted
        Tind GetApproximation(std::valarray<Tdata>& x, const Tind& t, size_t order = 1) const
        {
                MDS_ASSERT(size(), EXCEPTION,
                        "[result_manager] empty container in GetApproximation([valarray<Tdata>&,] const Tind&, size_t)");
        
                MDS_ASSERT(x.size() >= (begin()->second).size(), EXCEPTION,
                        "[result_manager] insufficient size of valarray in GetApproximation([valarray<Tdata>&,] const Tind&, size_t)");
        
                typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::const_iterator exactFind = find(t);

                Tind returnValue;

                size_t cnt, dim = (begin()->second).size(); 
                
                if (exactFind != end())
                {
                        MDS_ASSERT(count(t) == 1, EXCEPTION,
                                "[result_manager] more than one result in GetApproximation([valarray<Tdata>&,] const Tind&, size_t)");
                                
                        for (cnt = 0; cnt < dim; cnt++)
                                x[cnt] = (exactFind->second)[cnt];

                        returnValue = t - t;
                }
                else
                {
                        typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::const_iterator lowerKey = lower_bound(t);
                        typename std::multimap<Tind, std::valarray<Tdata>, Cmp, A>::const_iterator upperKey = lowerKey;
                        
                        if (upperKey == begin())
                        {
                                MDS_ASSERT(count(upperKey->first) == 1, EXCEPTION,
                                        "[result_manager] more than one result in GetApproximation([valarray<Tdata>&,] const Tind&, size_t)");
                                        
                                for (cnt = 0; cnt < dim; cnt++)
                                        x[cnt] = (begin()->second)[cnt];
                                
                                returnValue = begin()->first - t;
                        }
                        else
                        {
                                lowerKey--;

                                if (upperKey == end())
                                {
                                        MDS_ASSERT(count(lowerKey->first) == 1, EXCEPTION,
                                                "[result_manager] more than one result in GetApproximation([valarray<Tdata>&,] const Tind&, size_t)");
                                                
                                        for (cnt = 0; cnt < dim; cnt++)
                                                x[cnt] = (lowerKey->second)[cnt];
                                
                                        returnValue = lowerKey->first - t;
                                }
                                else
                                {
                                        MDS_ASSERT(order <= 1, EXCEPTION,
                                                "[result_manager] unsupported order of approximation in GetApproximation([valarray<Tdata>&,] const Tind&, size_t)");
                                                
                                        switch (order)
                                        {
                                        case 0:  // bang-bang approximation (by left value)
                                                MDS_ASSERT(count(lowerKey->first) == 1, EXCEPTION,
                                                        "[result_manager] more than one result in GetApproximation([valarray<Tdata>&,] const Tind&, size_t)");
                                                        
                                                for (cnt = 0; cnt < dim; cnt++)
                                                        x[cnt] = (lowerKey->second)[cnt];
                                                
                                                returnValue = t - lowerKey->first;
                                                
                                                break;

                                        case 1:  // linear approximation
                                                MDS_ASSERT(count(upperKey->first) == 1 && count(lowerKey->first) == 1, EXCEPTION,
                                                        "[result_manager] more than one result in GetApproximation([valarray<Tdata>&,] const Tind&, size_t)");
                                                        
                                                Tdata temp( (t - lowerKey->first) / (upperKey->first - lowerKey->first) );
                                                
                                                for (cnt = 0; cnt < dim; cnt++)
                                                        x[cnt] = (lowerKey->second + (upperKey->second - lowerKey->second) * temp)[cnt];
                                                
                                                if (t - lowerKey->first < upperKey->first - t) returnValue = t - lowerKey->first;
                                                else returnValue = upperKey->first - t;
                                                
                                                break;
                                        }
                                }
                        }
                }
                
                return returnValue;
        }

// approximation with order (0:bang-bang,1:linear)
        std::valarray<Tdata> GetApproximation(const Tind& t, size_t order = 1) const
        {
                MDS_ASSERT(size(), EXCEPTION,
                        "[result_manager] empty container in GetApproximation(const Tind&, size_t)");
        
                std::valarray<Tdata> v((begin()->second).size());

                GetApproximation(v, t, order);

                return v;
        }

};


#endif  
